"""TcEx Framework Module"""

# standard library
import json
import logging
import os
import re
from base64 import b64decode
from pathlib import Path

# third-party
from pydantic import BaseModel, Extra, ValidationError

# first-party
from tcex.app.config.install_json import InstallJson
from tcex.input.field_type import Sensitive
from tcex.input.model.advanced_request_model import AdvancedRequestModel
from tcex.input.model.app_external_model import AppExternalModel
from tcex.input.model.common_advanced_model import CommonAdvancedModel
from tcex.input.model.common_model import CommonModel
from tcex.input.model.model_map import feature_map, runtime_level_map, tc_action_map
from tcex.input.model.module_app_model import ModuleAppModel
from tcex.input.model.module_requests_session_model import ModuleRequestsSessionModel
from tcex.input.model.path_model import PathModel
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property
from tcex.registry import registry
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore

# define JSON encoders
json_encoders = {Sensitive: lambda v: str(v)}


def input_model(models: list) -> CommonModel | CommonAdvancedModel:
    """Return Input Model."""

    class InputModel(*models):
        """Input Model"""

        # [legacy] if True, the App should get it's inputs from secure params (redis)
        # supported runtimeLevel: [Organization, Playbook]
        tc_secure_params: bool = False

        # the user id of the one executing the App
        # supported runtimeLevel: [Organization, Playbook]
        tc_user_id: int | None

        class Config:
            """DataModel Config"""

            extra = Extra.allow
            validate_assignment = True
            json_encoders = json_encoders

    return InputModel


class Input:
    """Module to handle inputs for all App types."""

    def __init__(self, config: dict | None = None, config_file: str | None = None):
        """Initialize instance properties."""

        self.config = config
        self.config_file = config_file

        # properties
        self._models = []
        self.ij = InstallJson()
        self.log = _logger
        self.util = Util()

    def _load_config_file(self):
        """Load config file params provided passed to inputs."""
        # default file contents
        file_content = {}
        if self.config_file is None:
            return file_content

        # self.config_file should be a fully qualified file name
        fqfn = Path(self.config_file)
        if not fqfn.is_file():
            self.log.error(
                'feature=inputs, event=load-config-file, '
                f'exception=file-not-found, filename={fqfn.name}'
            )
            return file_content

        # read file contents
        try:
            # read encrypted file from "in" directory
            with fqfn.open(mode='rb') as fh:
                return json.load(fh)
        except Exception:  # pragma: no cover
            self.log.exception(f'feature=inputs, event=config-parse-failure, filename={fqfn.name}')
            return file_content

    def _load_file_params(self):
        """Load file params provided by the core platform."""
        # default file contents
        file_content = {}

        tc_app_param_file = os.getenv('TC_APP_PARAM_FILE')
        tc_app_param_key = os.getenv('TC_APP_PARAM_KEY')
        if tc_app_param_file and tc_app_param_key:
            # tc_app_param_file is a fully qualified file name
            fqfn = Path(tc_app_param_file)
            if not fqfn.is_file():  # pragma: no cover
                self.log.error(
                    'feature=inputs, event=load-file-params, '
                    f'exception=file-not-found, filename={fqfn.name}'
                )
                return file_content

            # read file contents
            try:
                # read encrypted file from "in" directory
                with fqfn.open(mode='rb') as fh:
                    encrypted_contents = fh.read()
            except Exception:  # pragma: no cover
                self.log.exception(
                    f'feature=inputs, event=config-parse-failure, filename={fqfn.name}'
                )
                return file_content

            # decrypt file contents
            try:
                file_content = json.loads(
                    self.util.decrypt_aes_cbc(tc_app_param_key, encrypted_contents).decode()
                )

                # delete file
                fqfn.unlink()
            except Exception:  # pragma: no cover
                self.log.exception(
                    f'feature=inputs, event=config-decryption-failure, filename={fqfn.name}'
                )

        return file_content

    def add_model(self, model: type[BaseModel]):
        """Add additional input model."""
        if model:
            self._models.insert(0, model)

        # clear cache for data property
        if 'model' in self.__dict__:
            del self.__dict__['model']

        # add App level model based on special "tc_action" input. this field
        # doesn't exist on the common model, but can be added by the App.
        # the tc_action can never be a variable (choice input).
        if tc_action := self.contents.get('tc_action'):
            self._models.extend(tc_action_map.get(tc_action, []))

        # force data model to load so that validation is done at this EXACT point
        _ = self.model

    @cached_property
    def contents(self) -> dict:
        """Return contents of inputs from all locations."""
        _contents = {}

        # config
        if isinstance(self.config, dict):
            _contents.update(self.config)

        # config file
        _contents.update(self._load_config_file())

        # file params
        _contents.update(self._load_file_params())

        return _contents

    @cached_property
    def contents_resolved(self) -> dict:
        """Resolve all file, keychain, playbook, and text variables.

        Job, Playbook, and Service Apps call can have a tc-variable, but only
        Playbook Apps will have a playbook variable.
        """
        _inputs = self.contents.copy()

        # support external Apps that don't have an install.json
        if not self.ij.fqfn.is_file():  # pragma: no cover
            return _inputs

        for name, value in _inputs.items():
            if name == 'tc_playbook_out_variables':
                # for services, tc_playbook_out_variables contains the name of the expected outputs.
                # If we don't skip this, we'll try to resolve the value (e.g.
                # #Trigger:334:example.service_input!String)
                # 1. this won't work for services
                # 2. service configs will never have playbook variables
                continue

            if self.util.is_tc_variable(value):  # only matches threatconnect variables
                _inputs[name] = self.resolve_variable(variable=value)
            elif self.ij.model.is_playbook_app:
                if isinstance(value, list):
                    # list could contain playbook variables, try to resolve the value
                    updated_value_array = []
                    for v in value:
                        if isinstance(v, str):
                            updated_value_array.append(registry.playbook.read.variable(v))
                        else:
                            # TODO: [high] does resolve variable need to be added here
                            updated_value_array.append(v)
                    _inputs[name] = updated_value_array
                elif self.util.is_playbook_variable(value):  # only matches playbook variables
                    # when using Bytes | String in App input model the value
                    # can be coerced to the wrong type. the BinaryVariable and
                    # StringVariable custom types allows for the validator in Binary
                    # and String types to raise a value error.
                    _inputs[name] = registry.playbook.read.variable(value)
                elif isinstance(value, str):
                    _inputs[name] = registry.playbook.read._read_embedded(value)  # noqa: SLF001
                else:
                    _inputs[name] = value
            else:
                for match in re.finditer(self.util.variable_tc_pattern, str(value)):
                    variable = match.group(0)  # the full variable pattern
                    if match.group('type').lower() == 'file':
                        v = '<file>'
                    else:
                        v = self.resolve_variable(variable=variable)
                    _inputs[name] = value.replace(variable, v)

        # update contents
        self.contents_update(_inputs)
        return dict(sorted(_inputs.items()))

    # TODO: [high] - can this be replaced with a pydantic root validator?
    def contents_update(self, inputs: dict):
        """Update inputs provided by core to be of the proper value and type."""
        for name, value in inputs.items():
            # ThreatConnect params could be updated in the future to proper JSON format.
            # MultiChoice data should be represented as JSON array and Boolean values should be a
            # JSON boolean and not a string.
            param = self.ij.model.get_param(name)
            if param is None:
                # skip over "default" inputs not defined in the install.json file
                continue

            if param.type.lower() == 'multichoice' or param.allow_multiple:
                # update delimited value to an array for inputs that have type of MultiChoice
                if value is not None and not isinstance(value, list):
                    inputs[name] = value.split(self.ij.model.list_delimiter or '|')
            elif param.type == 'boolean' and isinstance(value, str):
                # convert boolean input that are passed in as a string ("true" -> True)
                inputs[name] = value.lower() == 'true'

    @cached_property
    def model(self) -> CommonAdvancedModel:
        """Return the Input Model."""
        return input_model(self.models)(**self.contents_resolved)  # type: ignore

    @cached_property
    def model_advanced_request(self) -> AdvancedRequestModel:
        """Return the Requests Session Model."""

        class _AdvancedRequestModel(AdvancedRequestModel, extra=Extra.ignore):
            """Model Definition for AdvancedRequestModel inputs ONLY."""

        return _AdvancedRequestModel(**self.contents)

    @cached_property
    def model_organization_unresolved(self) -> CommonModel:
        """Return the Input Model using contents (no resolved values)."""
        return input_model(self.models)(**self.contents)  # type: ignore

    @cached_property
    def model_path(self) -> PathModel:
        """Return the Requests Session Model."""

        class _PathModel(PathModel, extra=Extra.ignore):
            """Model Definition for PathModel inputs ONLY."""

        return _PathModel(**self.contents)

    @cached_property
    def model_unresolved(self) -> CommonAdvancedModel:
        """Return full data model with no resolved values.

        This model has all inputs, including App inputs (e.g., tc_action), but
        any pb/tc variables will not be resolved in the model. The model is
        useful getting and process keyvalue variables that have mixed types.
        """
        return input_model(self.models)(**self.contents)  # type: ignore

    @cached_property
    def model_tc(self) -> CommonAdvancedModel:
        """Return data model for ThreatConnect specific inputs.

        This model does not have any App inputs (e.g., tc_action) and
        should only be used for accessing ThreatConnect specific inputs.
        """

        class _CommonAdvancedModel(CommonAdvancedModel, extra=Extra.ignore):
            """Model Definition for CommonAdvancedModel inputs ONLY."""

        return _CommonAdvancedModel(**self.contents)

    @cached_property
    def models(self) -> list:
        """Return all model for inputs."""
        # support external Apps that don't have an install.json
        if not self.ij.fqfn.is_file():
            return [AppExternalModel]

        # add all model for any supported features of the App
        for feature in self.ij.model.features:
            self._models.extend(feature_map.get(feature, []))

        # add all model based on the runtime level of the App
        rlm = runtime_level_map.get(self.ij.model.runtime_level.lower())
        if rlm is not None:
            self._models.append(rlm)

        return self._models

    # TODO: [low] is this needed or can Field value be set to tc_property=True?
    @cached_property
    def model_properties(self) -> set:
        """Return only defined properties from model (exclude additional)."""
        properties = set()
        for model in self.models:
            properties.update(model.schema().get('properties').keys())

        return properties

    @cached_property
    def module_app_model(self) -> ModuleAppModel:
        """Return the Module App Model."""
        return ModuleAppModel(**self.contents)

    @cached_property
    def module_requests_session_model(self) -> ModuleRequestsSessionModel:
        """Return the Module Requests Session Model."""
        return ModuleRequestsSessionModel(**self.contents)

    def resolve_variable(self, variable: str) -> bytes | str | Sensitive:
        """Resolve FILE/KEYCHAIN/TEXT variables.

        Feature: PLAT-2688

        Data Format:
        {
                "data": "value"
        }
        """
        match = re.match(Util().variable_tc_match, variable)
        if not match:
            ex_msg = f'Could not parse variable: {variable}'
            raise RuntimeError(ex_msg)

        key = match.group('key')
        provider = match.group('provider')
        type_ = match.group('type')

        # retrieve value from API
        data = None
        r = registry.session_tc.get(f'/internal/variable/runtime/{provider}/{key}')
        if r.ok:
            try:
                data = r.json().get('data')

                if type_.lower() == 'file':
                    data = b64decode(data)  # returns bytes
                elif type_.lower() == 'keychain':
                    data = Sensitive(data)
            except Exception as ex:
                ex_msg = (
                    f'Could not retrieve variable: provider={provider}, key={key}, type={type_}.'
                )
                raise RuntimeError(ex_msg) from ex
        else:
            ex_msg = f'Could not retrieve variable: provider={provider}, key={key}, type={type_}.'
            raise RuntimeError(ex_msg)

        return data

    @staticmethod
    def validation_exit_message(ex: ValidationError):
        """Format and return validation error message."""
        _exit_message = {}
        for err in ex.errors():
            # deduplicate error messages
            err_loc = ','.join([str(e) for e in err.get('loc')])
            err_msg = err.get('msg')
            _exit_message.setdefault(err_loc, [])
            if err_msg not in _exit_message[err_loc]:
                _exit_message[err_loc].append(err_msg)

        # format error messages
        _exit_message_list = ['Input validation errors']
        for loc, err_list in _exit_message.items():
            _exit_message_list.append(f"""{loc}: {', '.join(err_list)}""")

        # exit with error message
        return '\n'.join(_exit_message_list)



"""TcEx Framework Module"""

# standard library
import re
from collections.abc import Generator

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.app.config.install_json import InstallJson
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidInput, InvalidType
from tcex.registry import registry


class EditChoice(str):
    """EditChoice Field Type

    Used for Apps Param types:
    * EditChoice (pre-defined values)

    Unsupported:
    * Magic Variables (e.g. ${OWNERS})
    * EditChoice (dynamic values supported)
    """

    _allow_additional = False
    _value_transformations = None

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield cls.validate_type
        yield cls.modifier_strip
        yield cls.validate_valid_values

    @classmethod
    def modifier_strip(cls, value: str) -> str:
        """Modify value, stripping whitespace."""
        return value.strip()

    @classmethod
    def validate_type(cls, value: str, field: ModelField) -> str:
        """Raise exception if value is not a String type."""
        if not isinstance(value, str):
            raise InvalidType(
                field_name=field.name, expected_types='(str)', provided_type=type(value)
            )
        return value

    @classmethod
    def validate_valid_values(cls, value: str, field: ModelField) -> str:
        """Raise exception if value is not a String type."""
        if value == '':
            raise InvalidEmptyValue(field.name)

        # fix addressing PLAT-14751
        pattern = re.compile(r'^\${(users|user_group):.*?}$', re.IGNORECASE)
        if pattern.match(value):
            value = re.sub(r'^\${(users|user_group):|}$', '', value, flags=re.IGNORECASE)

        ij = InstallJson()
        param = ij.model.get_param(field.name)

        # TODO: [high] figure out a better way ...
        if param is None:
            # for a multichoice input field, pydantic prefixes the name with an underscore,
            # this breaks the lookup based on the "name" field in the install.json. Strip
            # the underscore to get the correct param name.
            param = ij.model.get_param(field.name.lstrip('_'))

        ti_utils = ThreatIntelUtil(registry.session_tc)
        valid_values = [] if param is None else param.valid_values
        _valid_values = ti_utils.resolve_variables(valid_values)
        for vv in _valid_values:
            if vv.lower() == value.lower():
                value = vv
                break
        else:
            if cls._allow_additional is False:
                raise InvalidInput(
                    field_name=field.name,
                    error=f'provided value {value} is not a valid value {_valid_values}',
                )

        if isinstance(cls._value_transformations, dict):
            value = cls._value_transformations.get(value, value)

        return value


def edit_choice(
    allow_additional: bool = False, value_transformations: dict[str, str] | None = None
) -> type[EditChoice]:
    """Return configured instance of String.

    :param allow_additional: Denotes whether this field will allow values that are not found in
    the field's valid values list in the install.json.
    :param value_transformations: dictionary that dictates how a choice should be transformed.
    Dictionary keys should be the field's valid values as defined in the install.json. Example:

    value_transformations: {'my_choice': 'My Choice'}

    If this field were to be initialized with 'my_choice', then the final value found in the input
    model would be 'My Choice'.
    """
    namespace = {
        '_value_transformations': value_transformations,
        '_allow_additional': allow_additional,
    }
    return type('CustomEditChoice', (EditChoice,), namespace)



"""TcEx Framework Module"""

# standard library
import gzip
import json
import logging
import math
import re
import time

# third-party
from requests import Session

# first-party
from tcex.exit.error_code import handle_error
from tcex.input.input import Input
from tcex.logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class BatchSubmit:
    """ThreatConnect Batch Import Module"""

    def __init__(
        self,
        inputs: Input,
        session_tc: Session,
        owner: str,
        action: str = 'Create',
        attribute_write_type: str = 'Replace',
        halt_on_error: bool = True,
        playbook_triggers_enabled: bool = False,
        tag_write_type: str = 'Replace',
        security_label_write_type: str = 'Replace',
    ):
        """Initialize instance properties.

        Args:
            inputs: The App inputs.
            session_tc: The ThreatConnect API session.
            owner: The ThreatConnect owner for Batch action.
            action: Action for the batch job ['Create', 'Delete'].
            attribute_write_type: Write type for Indicator attributes ['Append', 'Replace'].
            halt_on_error: If True any batch error will halt the batch job.
            playbook_triggers_enabled: Enables firing of playbooks.
            security_label_write_type: Write type for labels ['Append', 'Replace'].
            tag_write_type: Write type for tags ['Append', 'Replace'].
        """
        self.inputs = inputs
        self.session_tc = session_tc
        self._action = action
        self._attribute_write_type = attribute_write_type
        self._halt_on_error = halt_on_error
        self._owner = owner
        self._playbook_triggers_enabled = playbook_triggers_enabled
        self._security_label_write_type = security_label_write_type
        self._tag_write_type = tag_write_type

        # properties
        self._file_merge_mode = None
        self._hash_collision_mode = None
        self.log = _logger

        # global overrides on batch/file errors
        self._halt_on_batch_error = None
        self._halt_on_poll_error = None

        # default properties
        self._batch_data_count = None
        self._poll_interval = None
        self._poll_interval_times = []
        self._poll_timeout = 3600

    @property
    def _critical_failures(self) -> list[str]:  # pragma: no cover
        """Return Batch critical failure messages."""
        return [
            'Encountered an unexpected Exception while processing batch job',
            'would exceed the number of allowed indicators',
        ]

    @property
    def action(self) -> str:
        """Return batch action."""
        return self._action

    @action.setter
    def action(self, action):
        """Set batch action."""
        self._action = action

    @property
    def attribute_write_type(self) -> str:
        """Return batch attribute write type."""
        return self._attribute_write_type

    @attribute_write_type.setter
    def attribute_write_type(self, write_type: str):
        """Set batch attribute write type."""
        self._attribute_write_type = write_type

    def create_job(self, halt_on_error: bool = True) -> int | None:
        """Submit Batch request to ThreatConnect API.

        Args:
            halt_on_error: If True any exception will raise an error.

        Returns:
            int: The batch id from the API response.
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        try:
            r = self.session_tc.post('/v2/batch', json=self.settings)
        except Exception as e:
            handle_error(code=10505, message_values=[e], raise_error=halt_on_error)
            return None

        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            handle_error(
                code=10510,
                message_values=[r.status_code, r.text],
                raise_error=halt_on_error,
            )

        data = r.json()
        if data.get('status') != 'Success':
            handle_error(
                code=10510,
                message_values=[r.status_code, r.text],
                raise_error=halt_on_error,
            )

        self.log.debug(f'feature=batch, event=submit-job, status={data}')
        return data.get('data', {}).get('batchId')

    @property
    def error_codes(self) -> dict[str, str]:
        """Return static list of Batch error codes and short description"""
        return {
            '0x1001': 'General Error',
            '0x1002': 'Permission Error',
            '0x1003': 'JsonSyntax Error',
            '0x1004': 'Internal Error',
            '0x1005': 'Invalid Indicator Error',
            '0x1006': 'Invalid Group Error',
            '0x1007': 'Item Not Found Error',
            '0x1008': 'Indicator Limit Error',
            '0x1009': 'Association Error',
            '0x100A': 'Duplicate Item Error',
            '0x100B': 'File IO Error',
            '0x2001': 'Indicator Partial Loss Error',
            '0x2002': 'Group Partial Loss Error',
            '0x2003': 'File Hash Merge Error',
            '0x3001': 'File Hash Merge Error',
        }

    def errors(self, batch_id: int, halt_on_error: bool = True) -> list:
        """Retrieve Batch errors to ThreatConnect API.

        .. code-block:: javascript

            [{
                "errorReason": "Incident incident-001 has an invalid status.",
                "errorSource": "incident-001 is not valid."
            }, {
                "errorReason": "Incident incident-002 has an invalid status.",
                "errorSource":"incident-002 is not valid."
            }]

        Args:
            batch_id: The ID returned from the ThreatConnect API for the current batch job.
            halt_on_error: If True any exception will raise an error.

        Returns:
            list: A list of batch errors.
        """
        errors = []
        try:
            self.log.debug(f'feature=batch, event=retrieve-errors, batch-id={batch_id}')
            r = self.session_tc.get(f'/v2/batch/{batch_id}/errors')
            # API does not return correct content type
            if r.ok:
                errors = json.loads(r.text)
            # temporarily process errors to find "critical" errors.
            # FR in core to return error codes.
            for error in errors:
                error_reason = error.get('errorReason')
                for error_msg in self._critical_failures:
                    if re.findall(error_msg, error_reason):
                        handle_error(
                            code=10500,
                            message_values=[error_reason],
                            raise_error=halt_on_error,
                        )
        except Exception as e:
            handle_error(code=560, message_values=[e], raise_error=halt_on_error)

        return errors

    def file_merge_mode(self, value: str):
        """Set the file merge mode for the entire batch job.

        Args:
            value: A value of Distribute or Merge.
        """
        self._file_merge_mode = value

    @property
    def halt_on_error(self) -> bool:
        """Return batch halt on error setting."""
        return self._halt_on_error

    @halt_on_error.setter
    def halt_on_error(self, halt_on_error: bool):
        """Set batch halt on error setting."""
        self._halt_on_error = halt_on_error

    @property
    def halt_on_batch_error(self) -> bool:
        """Return halt on batch error value."""
        return self._halt_on_batch_error or False

    @halt_on_batch_error.setter
    def halt_on_batch_error(self, value: bool):
        """Set batch halt on batch error value."""
        if isinstance(value, bool):
            self._halt_on_batch_error = value

    @property
    def halt_on_poll_error(self) -> bool:
        """Return halt on poll error value."""
        return self._halt_on_poll_error or False

    @halt_on_poll_error.setter
    def halt_on_poll_error(self, value: bool):
        """Set batch halt on poll error value."""
        if isinstance(value, bool):
            self._halt_on_poll_error = value

    def hash_collision_mode(self, value: str):
        """Set the file hash collision mode for the entire batch job.

        Args:
            value: A value of Split, IgnoreIncoming, IgnoreExisting, FavorIncoming,
                and FavorExisting.
        """
        self._hash_collision_mode = value

    def poll(
        self,
        batch_id: int,
        retry_seconds: int | None = None,
        back_off: float | None = None,
        timeout: int | None = None,
        halt_on_error: bool = True,
    ) -> dict:
        """Poll Batch status to ThreatConnect API.

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "batchStatus": {
                        "id":3505,
                        "status":"Completed",
                        "errorCount":0,
                        "successCount":0,
                        "unprocessCount":0
                    }
                }
            }

        Args:
            batch_id: The ID returned from the ThreatConnect API for the current batch job.
            retry_seconds: The base number of seconds used for retries when job is not completed.
            back_off: A multiplier to use for backing off on
                each poll attempt when job has not completed.
            timeout: The number of seconds before the poll should timeout.
            halt_on_error: If True any exception will raise an error.

        Returns:
            dict: The batch status returned from the ThreatConnect API.
        """
        # check global setting for override
        if self.halt_on_poll_error is not None:
            halt_on_error = self.halt_on_poll_error

        # initial poll interval
        if self._poll_interval is None and self._batch_data_count is not None:
            # calculate poll_interval base off the number of entries in the batch data
            # with a minimum value of 5 seconds.
            self._poll_interval = max(math.ceil(self._batch_data_count / 300), 5)
        elif self._poll_interval is None:
            # if not able to calculate poll_interval default to 15 seconds
            self._poll_interval = 15

        # poll retry back_off factor
        poll_interval_back_off = float(2.5 if back_off is None else back_off)

        # poll retry seconds
        poll_retry_seconds = int(5 if retry_seconds is None else retry_seconds)

        # poll timeout
        timeout = self.poll_timeout if timeout is None else int(timeout)
        params = {'includeAdditional': 'true'}

        poll_count = 0
        poll_time_total = 0
        data = {}
        while True:
            poll_count += 1
            poll_time_total += self._poll_interval
            time.sleep(self._poll_interval)
            self.log.info(f'feature=batch, event=progress, poll-time={poll_time_total}')
            try:
                # retrieve job status
                r = self.session_tc.get(f'/v2/batch/{batch_id}', params=params)
                if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                    handle_error(
                        code=545,
                        message_values=[r.status_code, r.text],
                        raise_error=halt_on_error,
                    )
                    return data
                data = r.json()
                if data.get('status') != 'Success':
                    handle_error(
                        code=545,
                        message_values=[r.status_code, r.text],
                        raise_error=halt_on_error,
                    )
            except Exception as e:
                handle_error(code=540, message_values=[e], raise_error=halt_on_error)

            if data.get('data', {}).get('batchStatus', {}).get('status') == 'Completed':
                # store last 5 poll times to use in calculating average poll time
                modifier = poll_time_total * 0.7
                self._poll_interval_times = self._poll_interval_times[-4:] + [modifier]

                weights: list[float | int] = [1]
                poll_interval_time_weighted_sum = 0
                for poll_interval_time in self._poll_interval_times:
                    poll_interval_time_weighted_sum += poll_interval_time * weights[-1]
                    # weights will be [1, 1.5, 2.25, 3.375, 5.0625] for all 5 poll times depending
                    # on how many poll times are available.
                    weights.append(weights[-1] * 1.5)

                # pop off the last weight so its not added in to the sum
                weights.pop()

                # calculate the weighted average of the last 5 poll times
                self._poll_interval = math.floor(poll_interval_time_weighted_sum / sum(weights))

                if poll_count == 1:
                    # if completed on first poll, reduce poll interval.
                    self._poll_interval = self._poll_interval * 0.85

                self.log.debug(f'feature=batch, poll-time={poll_time_total}, status={data}')
                return data

            # update poll_interval for retry with max poll time of 20 seconds
            self._poll_interval = min(
                poll_retry_seconds + int(poll_count * poll_interval_back_off), 20
            )

            # time out poll to prevent App running indefinitely
            if poll_time_total >= timeout:
                handle_error(code=550, message_values=[timeout], raise_error=True)

    @property
    def poll_timeout(self) -> int:
        """Return current poll timeout value."""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, seconds: int):
        """Set the poll timeout value."""
        self._poll_timeout = int(seconds)

    @property
    def security_label_write_type(self) -> str:
        """Return batch security label write type."""
        return self._security_label_write_type

    @security_label_write_type.setter
    def security_label_write_type(self, write_type: str):
        """Set batch security label write type."""
        self._security_label_write_type = write_type

    @property
    def settings(self) -> dict[str, str]:
        """Return batch job settings."""
        _settings = {
            'action': self._action,
            'attributeWriteType': self.attribute_write_type,
            'haltOnError': str(self._halt_on_error).lower(),
            'owner': self._owner,
            'playbookTriggersEnabled': str(self._playbook_triggers_enabled).lower(),
            'securityLabelWriteType': self.security_label_write_type,
            'tagWriteType': self.tag_write_type,
            'version': 'V2',
        }
        if self._hash_collision_mode is not None:
            _settings['hashCollisionMode'] = self._hash_collision_mode
        if self._file_merge_mode is not None:
            _settings['fileMergeMode'] = self._file_merge_mode
        return _settings

    def submit(self, batch_filename: str, halt_on_error: bool = True) -> dict:
        """Submit Batch request to ThreatConnect API.

        Args:
            batch_filename: The filename for the batch JSON file.
            halt_on_error: If True the process should halt if any errors are encountered.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        # content = gzip.open(batch_filename, 'rt').read()
        with gzip.open(batch_filename, 'rt') as fh:
            content = fh.read()

            # check global setting for override
            if self.halt_on_batch_error is not None:
                halt_on_error = self.halt_on_batch_error

            files = (('config', json.dumps(self.settings)), ('content', content))
            params = {'includeAdditional': 'true'}
            try:
                r = self.session_tc.post('/v2/batch/createAndUpload', files=files, params=params)
                if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                    handle_error(
                        code=10510,
                        message_values=[r.status_code, r.text],
                        raise_error=halt_on_error,
                    )
                return r.json()
            except Exception as e:
                handle_error(code=10505, message_values=[e], raise_error=halt_on_error)

        return {}

    def submit_data(
        self, batch_id: int, content: dict | str, halt_on_error: bool = True
    ) -> dict | None:
        """Submit Batch request to ThreatConnect API.

        Args:
            batch_id: The batch id of the current job.
            content: The dict of groups and indicator data.
            halt_on_error (bool = True): If True the process
                should halt if any errors are encountered.

        Returns:
            dict: The response data
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        # TC Core requires the header to be application/octet-stream
        headers = {'Content-Type': 'application/octet-stream'}
        try:
            if isinstance(content, dict):
                content = json.dumps(content)

            r = self.session_tc.post(f'/v2/batch/{batch_id}', headers=headers, data=content)
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                handle_error(
                    code=10525,
                    message_values=[r.status_code, r.text],
                    raise_error=halt_on_error,
                )
            return r.json()
        except Exception as e:
            handle_error(code=10520, message_values=[e], raise_error=halt_on_error)

        return None

    @property
    def tag_write_type(self) -> str:
        """Return batch tag write type."""
        return self._tag_write_type

    @tag_write_type.setter
    def tag_write_type(self, write_type: str):
        """Set batch tag write type."""
        self._tag_write_type = write_type



"""TcEx Framework Module"""

# standard library
from typing import Literal

from .render_panel import RenderPanel
from .render_prompt import RenderPrompt
from .render_table import RenderTable

AlignMethod = Literal['left', 'center', 'right']


class Render:
    """Render CLI Output"""

    accent = 'dark_orange'
    accent2 = 'dodger_blue2'
    panel = RenderPanel
    prompt = RenderPrompt
    table = RenderTable
    title_align: AlignMethod = 'left'



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class IntelRequirementFilter(FilterABC):
    """Filter Object for IntelRequirements"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.INTEL_REQUIREMENTS.value

    def category(self, operator: Enum, category: list | str):
        """Filter Category based on **category** keyword.

        Args:
            operator: The operator enum for the filter.
            category: The subtype of the intel requirement.
        """
        if isinstance(category, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('category', operator, category, TqlType.STRING)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the requirement was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_tag(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasTag', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    @property
    def has_victim_asset(self):
        """Return **VictimAssetFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset_filter import VictimAssetFilter

        victim_assets = VictimAssetFilter(Tql())
        self._tql.add_filter('hasVictimAsset', TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)
        return victim_assets

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter Bucket ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The bucket ID of the intel requirement.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def id_bucket(self, operator: Enum, id_bucket: int | list):
        """Filter Bucket ID based on **idBucket** keyword.

        Args:
            operator: The operator enum for the filter.
            id_bucket: The bucket ID of the intel requirement.
        """
        if isinstance(id_bucket, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('idBucket', operator, id_bucket, TqlType.INTEGER)

    def id_parent(self, operator: Enum, id_parent: int | list):
        """Filter ID Bucket Parent based on **idParent** keyword.

        Args:
            operator: The operator enum for the filter.
            id_parent: The ID of the parent intel requirement.
        """
        if isinstance(id_parent, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('idParent', operator, id_parent, TqlType.INTEGER)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date the requirement was last modified.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the intel requirement.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the intel requirement.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def requirement(self, operator: Enum, requirement: list | str):
        """Filter Requirement based on **requirement** keyword.

        Args:
            operator: The operator enum for the filter.
            requirement: The Requirement of the intel requirement.
        """
        if isinstance(requirement, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('requirement', operator, requirement, TqlType.STRING)

    def subtype(self, operator: Enum, subtype: list | str):
        """Filter Subtype based on **subtype** keyword.

        Args:
            operator: The operator enum for the filter.
            subtype: The subtype of the intel requirement.
        """
        if isinstance(subtype, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('subtype', operator, subtype, TqlType.STRING)

    def tag(self, operator: Enum, tag: list | str):
        """Filter Tag based on **tag** keyword.

        Args:
            operator: The operator enum for the filter.
            tag: The name of a tag applied to the group.
        """
        if isinstance(tag, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('tag', operator, tag, TqlType.STRING)

    def unique_id(self, operator: Enum, unique_id: list | str):
        """Filter Unique ID based on **uniqueId** keyword.

        Args:
            operator: The operator enum for the filter.
            unique_id: The unique ID of the intel requirement.
        """
        if isinstance(unique_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('uniqueId', operator, unique_id, TqlType.STRING)



"""TcEx Framework Module"""

# third-party
from requests import Session

# first-party
from tcex.api.tc.v3.security.exclusion_lists.exclusion_list import ExclusionList, ExclusionLists
from tcex.api.tc.v3.security.owner_roles.owner_role import OwnerRole, OwnerRoles
from tcex.api.tc.v3.security.owners.owner import Owner, Owners
from tcex.api.tc.v3.security.system_roles.system_role import SystemRole, SystemRoles
from tcex.api.tc.v3.security.user_groups.user_group import UserGroup, UserGroups
from tcex.api.tc.v3.security.users.user import User, Users


class Security:
    """Security

    Args:
        session: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, session: Session):
        """Initialize instance properties."""
        self.session = session

    def owner_role(self, **kwargs) -> OwnerRole:
        """Return a instance of Owner Role object."""
        return OwnerRole(session=self.session, **kwargs)

    def owner_roles(self, **kwargs) -> OwnerRoles:
        """Return a instance of Owner Roles object."""
        return OwnerRoles(session=self.session, **kwargs)

    def owner(self, **kwargs) -> Owner:
        """Return a instance of Owner object."""
        return Owner(session=self.session, **kwargs)

    def owners(self, **kwargs) -> Owners:
        """Return a instance of Owners object."""
        return Owners(session=self.session, **kwargs)

    def system_role(self, **kwargs) -> SystemRole:
        """Return a instance of System Role object."""
        return SystemRole(session=self.session, **kwargs)

    def system_roles(self, **kwargs) -> SystemRoles:
        """Return a instance of System Roles object."""
        return SystemRoles(session=self.session, **kwargs)

    def user_group(self, **kwargs) -> UserGroup:
        """Return a instance of User Group object."""
        return UserGroup(session=self.session, **kwargs)

    def user_groups(self, **kwargs) -> UserGroups:
        """Return a instance of User Group object."""
        return UserGroups(session=self.session, **kwargs)

    def exclusion_list(self, **kwargs) -> ExclusionList:
        """Return a instance of User Group object."""
        return ExclusionList(session=self.session, **kwargs)

    def exclusion_lists(self, **kwargs) -> ExclusionLists:
        """Return a instance of User Group object."""
        return ExclusionLists(session=self.session, **kwargs)

    def user(self, **kwargs) -> User:
        """Return a instance of User object."""
        return User(session=self.session, **kwargs)

    def users(self, **kwargs) -> Users:
        """Return a instance of Users object."""
        return Users(session=self.session, **kwargs)



"""TcEx Framework Module"""

# third-party
from redis import Redis

from .key_value_abc import KeyValueABC


class KeyValueRedis(KeyValueABC):
    """TcEx Key Value Redis Module.

    Args:
        redis_client (redis.Client): An instance of redis client.
    """

    def __init__(self, redis_client: Redis):
        """Initialize the Class properties."""
        self.redis_client = redis_client

        # properties
        self.kv_type = 'redis'

    def create(self, context: str, key: str, value: bytes | str) -> int:
        """Create key/value pair in Redis.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.
            value: The value for the kv pair in Redis.

        Returns:
            str: The number of fields that were added.
        """
        return self.redis_client.hset(context, key, value)  # type: ignore

    def delete(self, context: str, key: str) -> int:
        """Alias for hdel method.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        return self.redis_client.hdel(context, key)  # type: ignore

    def get_all(self, context: str) -> dict[str, bytes | str | None]:
        """Return the contents for a given context.

        Args:
            context: the context to return
        """
        return self.hgetall(context)

    def hgetall(self, context: str) -> dict[str, bytes | str | None]:
        """Read data from Redis for the current context.

        Args:
            context: A specific context for the create.

        Returns:
            list: The response data from Redis.
        """
        return self.redis_client.hgetall(context)  # type: ignore

    def read(self, context: str, key: str) -> bytes | str | None:
        """Read data from Redis for the provided key.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            str: The response data from Redis.
        """
        return self.hget(context, key)

    def hget(self, context: str, key: str) -> bytes | str | None:
        """Read data from redis for the provided key.

        This method will *not* convert the retrieved data (like read() does).

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            bytes | None: the raw value from redis, if any
        """
        return self.redis_client.hget(context, key)  # type: ignore



"""TcEx Framework Module"""

# standard library
from typing import ClassVar

# third-party
from pydantic import create_model, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_type.tc_entity import TCEntity

CASE_MANAGEMENT_TYPES = [
    'artifact',
    'case',
    'task',
    'artifact type',
    'notes',
    'workflow event',
    'workflow template',
]


class CaseManagementEntity(TCEntity):
    """Case Management Entity Field (Model) Type"""

    case_management_types: ClassVar[list[str]] = CASE_MANAGEMENT_TYPES

    @validator('type', allow_reuse=True)
    @classmethod
    def is_empty(cls, value: str, field: ModelField) -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @validator('type', allow_reuse=True)
    @classmethod
    def is_type(cls, value: str, field: ModelField) -> str:
        """Validate that the entity is of Indicator type."""
        if value.lower() not in [i.lower() for i in cls.case_management_types]:
            raise InvalidEntityType(
                field_name=field.name, entity_type=str(cls.case_management_types), value=value
            )
        return value


def case_management_entity(
    case_management_types: list[str], model_name: str = 'CustomCaseManagementEntity'
) -> type[CaseManagementEntity]:
    """Dynamically create a Case Management Entity model."""
    return create_model(
        model_name,
        case_management_types=(ClassVar[list[str]], case_management_types),
        __base__=CaseManagementEntity,
    )


ArtifactEntity = case_management_entity(case_management_types=['Artifact'])
CaseEntity = case_management_entity(case_management_types=['Case'])
TaskEntity = case_management_entity(case_management_types=['Task'])
ArtifactTypeEntity = case_management_entity(case_management_types=['Artifact Type'])
NoteEntity = case_management_entity(case_management_types=['Note'])
WorkflowEventEntity = case_management_entity(case_management_types=['Workflow Event'])
WorkflowTemplateEntity = case_management_entity(case_management_types=['Workflow Template'])



"""TcEx Framework Module"""

# first-party
from tcex.app.decorator.benchmark import Benchmark
from tcex.app.decorator.debug import Debug
from tcex.app.decorator.fail_on_output import FailOnOutput
from tcex.app.decorator.on_exception import OnException
from tcex.app.decorator.on_success import OnSuccess
from tcex.app.decorator.output import Output

__all__ = ['Benchmark', 'Debug', 'FailOnOutput', 'OnException', 'OnSuccess', 'Output']



"""TcEx Framework Module"""

# third-party
from pydantic import Extra, Field

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class IntelReqTypeModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Intel Requirement Type Model',
    validate_assignment=True,
):
    """Model Definition"""

    id: int | None = Field(  # pyright: ignore [reportGeneralTypeIssues]
        None,
        methods=['POST', 'PUT'],
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        description='Name of the Intel Requirement Type.',
        read_only=True,
        title='name',
    )
    description: str | None = Field(
        None,
        description='Description of the Intel Requirement Type.',
        read_only=True,
        title='description',
    )



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING, Any, cast

# third-party
import wrapt

# first-party
from tcex.exit import ExitCode
from tcex.tcex import TcEx

if TYPE_CHECKING:
    # standard library
    from collections.abc import Callable


class OnException:
    """Set exit message on failed execution.

    This decorator will catch the generic "Exception" error, log the supplied error message, set
    the "exit_message", and exit the App with an exit code of 1.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @OnException(exit_msg='Failed to process JSON data.')
        def my_method(json_data):
            json.dumps(json_data)

    Args:
        exit_msg (str): The message to send to exit method.
        exit_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        write_output (boolean): default True.
            If enabled, will call app.write_output() when an exception is raised.
    """

    def __init__(
        self,
        exit_msg: str | None = None,
        exit_enabled: bool | str = True,
        write_output: bool = True,
    ):
        """Initialize instance properties"""
        self.exit_enabled = exit_enabled
        self.exit_msg = exit_msg or 'An exception has been caught. See the logs for more details.'
        self.write_output = write_output

    @wrapt.decorator
    def __call__(self, *wrapped_args) -> Any:  # noqa: D417
        """Implement __call__ function for decorator.

        Args:
            wrapped: The wrapped function which in turns
                needs to be called by your wrapper function.
            instance: The object to which the wrapped
                function was bound when it was called.
            args: The list of positional arguments supplied
                when the decorated function was called.
            kwargs: The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """
        # using wrapped args to support typing hints in PyRight
        wrapped: Callable = wrapped_args[0]
        app: Any = wrapped_args[1]
        args: list = wrapped_args[2] if len(wrapped_args) > 1 else []
        kwargs: dict = wrapped_args[3] if len(wrapped_args) > 2 else {}  # noqa: PLR2004

        def exception() -> Any:
            """Call the function and handle any exception."""
            tcex = cast(TcEx, app.tcex)

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.exit_enabled
            if not isinstance(self.exit_enabled, bool):
                enabled = getattr(tcex.inputs.model, self.exit_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    ex_msg = 'The exit_enabled value must be a boolean or resolved to bool.'
                    raise RuntimeError(ex_msg)  # noqa: TRY004
                tcex.log.debug(f'Fail enabled is {enabled} ({self.exit_enabled}).')

            try:
                return wrapped(*args, **kwargs)
            except Exception:
                tcex.log.exception('An exception has been caught.')
                app.exit_message = self.exit_msg  # for test cases
                if enabled:
                    if self.write_output:
                        tcex.app.playbook.output.process()
                        if hasattr(app, 'write_output'):
                            app.write_output()
                    tcex.exit.exit(ExitCode.FAILURE, self.exit_msg)
            return None

        return exception()



"""TcEx Framework Module"""

# standard library
import contextlib
import json
import logging
from functools import lru_cache
from typing import TYPE_CHECKING
from urllib.parse import unquote

# third-party
from requests import Response

# first-party
from tcex.api.tc.v2.threat_intelligence.tcex_ti_tc_request import TiTcRequest
from tcex.exit.error_code import TcExErrorCode
from tcex.logger.trace_logger import TraceLogger
from tcex.util import Util

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import (
        ThreatIntelligence,  # CIRCULAR-IMPORT
    )


# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class Mapping:
    """Common API calls for for Indicators/SecurityLabels/Groups and Victims"""

    def __init__(
        self,
        ti: 'ThreatIntelligence',
        main_type,
        api_type,
        sub_type,
        api_entity,
        api_branch,
        owner,
    ):
        """Initialize instance properties"""
        self._api_branch = api_branch
        self._api_entity = api_entity
        self._api_sub_type = sub_type
        self._api_type = api_type
        self._owner = owner
        self._session = ti.session_tc
        self.ti = ti
        self._type = main_type

        # properties
        self._data = {}
        self.log = _logger
        self.util = Util()
        self._tc_requests = TiTcRequest(ti.session_tc)
        self._unique_id = None

    @property
    @lru_cache  # noqa: B019
    def _error_codes(self) -> TcExErrorCode:
        """Return TcEx error codes."""
        return TcExErrorCode()

    @property
    def _excluded_properties(self):
        """Return a list of properties to exclude when creating the child Class."""
        return ['tcex', 'kwargs', 'api_endpoint']

    def _handle_error(
        self, code: int, message_values: list | None = None, raise_error: bool = True
    ):
        """Raise RuntimeError

        Args:
            code: The error code from API or SDK.
            message_values: The error message from API or SDK.
            raise_error: Raise a Runtime error. Defaults to True.

        Raises:
            RuntimeError: Raised a defined error.
        """
        try:
            if message_values is None:
                message_values = []
            message = self._error_codes.message(code).format(*message_values)
            self.log.error(f'Error code: {code}, {message}')
        except AttributeError as ex:
            self.log.exception(f'Incorrect error code provided ({code}).')
            ex_msg = 'Generic Failure, see logs for more details.'
            raise RuntimeError(100, ex_msg) from ex
        except IndexError as ex:
            self.log.exception(
                f'Incorrect message values provided for error code {code} ({message_values}).'
            )
            ex_msg = 'Generic Failure, see logs for more details.'
            raise RuntimeError(100, ex_msg) from ex
        if raise_error:
            raise RuntimeError(code, message)

    @property
    def as_entity(self):  # pragma: no cover
        """Return the object as an entity."""
        ex_msg = 'Child class must implement this method.'
        raise NotImplementedError(ex_msg)

    @property
    def as_dict(self):
        """Return the dict representation of the case management object."""
        properties = vars(self)
        as_dict = {}
        for key, value in properties.items():
            key_ = key.lstrip('_')
            if key_ in self._excluded_properties:
                continue
            value_ = value
            with contextlib.suppress(AttributeError):
                value_ = value.as_dict
            if value_ is None:
                continue
            as_dict[key_] = value_
        if not as_dict:
            return None
        return as_dict

    @property
    def type(self):
        """Return main type."""
        return self._type

    @property
    def api_sub_type(self):
        """Return sub type."""
        return self._api_sub_type

    @property
    def unique_id(self):
        """Return unique id."""
        return self._unique_id

    @property
    def tc_requests(self):
        """Return tc request object."""
        return self._tc_requests

    @property
    def api_type(self):
        """Return api type."""
        return self._api_type

    @property
    def api_branch(self):
        """Return api entity."""
        return self._api_branch

    @property
    def api_entity(self):
        """Return api entity."""
        return self._api_entity

    @property
    def owner(self):
        """Return unique id."""
        return self._owner

    @owner.setter
    def owner(self, owner):
        self._owner = owner

    @api_branch.setter
    def api_branch(self, api_branch):
        """Set the Api Entity."""
        self._api_branch = api_branch

    @api_entity.setter
    def api_entity(self, api_entity):
        """Set the Api Entity."""
        self._api_entity = api_entity

    @api_type.setter
    def api_type(self, api_type):
        """Set the Api Type."""
        self._api_type = api_type

    @tc_requests.setter
    def tc_requests(self, tc_requests):
        """Set the Tc Request Object."""
        self._tc_requests = tc_requests

    @api_sub_type.setter
    def api_sub_type(self, sub_type):
        """Set the Api Sub Type."""
        self._api_sub_type = sub_type

    @unique_id.setter
    def unique_id(self, unique_id):
        """Set the Unique Id."""
        self._unique_id = unique_id

    @property
    def data(self):
        """Return data."""
        return self._data

    @data.setter
    def data(self, data):
        """Set the data."""
        self._data = data
        if not self.unique_id:
            self._set_unique_id(data)

    def set(self, **kwargs):
        """Provide a generic way to update the attributes of a TC data object."""
        for arg, value in kwargs.items():
            if hasattr(self, 'add_key_value'):
                self.add_key_value(arg, value)  # type: ignore
            else:
                self._data[arg] = value

    def create(self):
        """Create the Indicator/Group/Victim or Security Label given Owner."""
        if not self.can_create():
            self._handle_error(920, [self.type])

        response = self.tc_requests.create(self.api_type, self.api_branch, self._data, self.owner)

        if self.tc_requests.success(response):
            self._set_unique_id(response.json().get('data').get(self.api_entity))

        return response

    def delete(self):
        """Delete the Indicator/Group/Victim or Security Label."""
        if not self.can_delete():
            self._handle_error(915, [self.type])

        return self.tc_requests.delete(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def update(self):
        """Update the Indicator/Group/Victim or Security Label."""
        if not self.can_update():
            self._handle_error(905, [self.type])

        return self.tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, self._data, owner=self.owner
        )

    def single(self, filters=None, params=None):
        """Get the Indicator/Group/Victim or Security Label."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.single(
            self.api_type,
            self.api_branch,
            self.unique_id,
            filters=filters,
            owner=self.owner,
            params=params,
        )

    def many(self, filters=None, params=None):
        """Get the Indicator/Group/Victim or Security Labels."""
        yield from self.tc_requests.many(
            self.api_type,
            self.api_branch,
            self.api_entity,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def request(self, result_limit, result_start, filters=None, params=None):
        """Get the Indicator/Group/Victim or Security Labels."""
        return self.tc_requests.request(
            self.api_type,
            self.api_branch,
            result_limit,
            result_start,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def tags(self, filters=None, params=None):
        """Get the tags from a Indicator/Group/Victim/Security Labels."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        yield from self.tc_requests.tags(
            self.api_type,
            self.api_branch,
            self.unique_id,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def tag(self, name, action='ADD', params=None):
        """Add a tag to a Indicator/Group/Victim/Security Label."""
        if not name:
            self._handle_error(925, ['name', 'tag', 'name', 'name', name])

        if not self.can_update():
            self._handle_error(910, [self.type])

        if action in ['GET', 'ADD', 'DELETE']:
            return self.tc_requests.tag(
                self.api_type,
                self.api_branch,
                self.unique_id,
                name,
                action=action,
                owner=self.owner,
                params=params,
            )
        self._handle_error(925, ['action', 'tag', 'action', 'action', action])
        return None

    def add_tag(self, name):
        """Add a tag to a Indicator/Group/Victim/Security Label."""
        return self.tag(name, action='ADD')

    def get_tag(self, name, params=None):
        """Get a tag from a Indicator/Group/Victim/Security Label."""
        return self.tag(name, action='GET', params=params)

    def delete_tag(self, name):
        """Delete a tag from a Indicator/Group/Victim/Security Label."""

        return self.tag(name, action='DELETE')

    def labels(self, filters=None, params=None):
        """Get the security labels from a Indicator/Group/Victim."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        yield from self.tc_requests.labels(
            self.api_type,
            self.api_branch,
            self.unique_id,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def label(self, label, action='ADD', params=None) -> Response | None:
        """Add a Security Label to a Indicator/Group or Victim."""

        if params is None:
            params = {}

        if not label:
            self._handle_error(925, ['label', 'Security Label', 'label', 'label', label])

        if not self.can_update():
            self._handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                label,
                owner=self.owner,
                params=params,
            )

        if action == 'ADD':
            return self.tc_requests.add_label(
                self.api_type, self.api_branch, self.unique_id, label, owner=self.owner
            )

        if action == 'DELETE':
            return self.tc_requests.delete_label(
                self.api_type, self.api_branch, self.unique_id, label, owner=self.owner
            )

        self._handle_error(925, ['action', 'label', 'action', 'action', action])
        return None

    def add_label(self, label):
        """Add a label to a Indicator/Group/Victim."""
        return self.label(label, action='ADD')

    def get_label(self, label, params=None):
        """Get a security label from a Indicator/Group/Victim."""
        return self.label(label, action='GET', params=params)

    def delete_label(self, label):
        """Delete a security label from a Indicator/Group/Victim."""
        return self.label(label, action='DELETE')

    def indicator_associations(self, params=None):
        """Get the indicator association from a Indicator/Group/Victim."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        if params is None:
            params = {}

        yield from self.tc_requests.indicator_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def group_associations(self, params=None):
        """Get the group association from a Indicator/Group/Victim."""
        if params is None:
            params = {}

        if not self.can_update():
            self._handle_error(910, [self.type])

        yield from self.tc_requests.group_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def victim_asset_associations(self, params=None):
        """Get the victim asset association from a Indicator/Group/Victim."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.victim_asset_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def indicator_associations_types(
        self, indicator_type, api_entity=None, api_branch=None, params=None
    ):
        """Get the indicator association from a Indicator/Group/Victim."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        target = self.ti.indicator(indicator_type)
        yield from self.tc_requests.indicator_associations_types(
            self.api_type,
            self.api_branch,
            self.unique_id,
            target,
            api_entity=api_entity,
            api_branch=api_branch,
            owner=self.owner,
            params=params,
        )

    def group_associations_types(self, group_type, api_entity=None, api_branch=None, params=None):
        """Get the group association from a Indicator/Group/Victim."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        target = self.ti.group(group_type)
        yield from self.tc_requests.group_associations_types(
            self.api_type,
            self.api_branch,
            self.unique_id,
            target,
            api_entity=api_entity,
            api_branch=api_branch,
            owner=self.owner,
            params=params,
        )

    def victim_asset_associations_type(self, victim_asset_type, params=None):
        """Get the victim association from a Indicator/Group/Victim."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.victim_asset_associations(
            self.api_type,
            self.api_branch,
            self.unique_id,
            victim_asset_type,
            owner=self.owner,
            params=params,
        )

    def add_association(self, target, api_type=None, api_branch=None, unique_id=None):
        """Add a association to a Indicator/Group/Victim."""
        api_type = api_type or target.api_type
        api_branch = api_branch or target.api_branch
        unique_id = unique_id or target.unique_id
        if not self.can_update():
            self._handle_error(910, [self.type])

        if not target.can_update():
            self._handle_error(910, [target.type])

        return self.tc_requests.add_association(
            self.api_type,
            self.api_branch,
            self.unique_id,
            api_type,
            api_branch,
            unique_id,
            owner=self.owner,
        )

    def delete_association(self, target, api_type=None, api_branch=None, unique_id=None):
        """Delete a association from a Indicator/Group/Victim."""
        api_type = api_type or target.api_type
        api_branch = api_branch or target.api_branch
        unique_id = unique_id or target.unique_id
        if not self.can_update():
            self._handle_error(910, [self.type])

        if not target.can_update():
            self._handle_error(910, [target.type])

        return self.tc_requests.delete_association(
            self.api_type,
            self.api_branch,
            self.unique_id,
            api_type,
            api_branch,
            unique_id,
            owner=self.owner,
        )

    def attributes(self, params=None):
        """Get the attributes from a Group/Indicator or Victim."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        yield from self.tc_requests.attributes(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def attribute(self, attribute_id, action='GET', params=None):
        """Get the attribute from a Group/Indicator or Victim."""
        if params is None:
            params = {}
        action = action.upper()
        if not self.can_update():
            self._handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_attribute(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                owner=self.owner,
                params=params,
            )

        if action == 'DELETE':
            return self.tc_requests.delete_attribute(
                self.api_type, self.api_branch, self.unique_id, attribute_id, owner=self.owner
            )

        self._handle_error(925, ['action', 'attribute', 'action', 'action', action])
        return None

    def add_attribute(
        self, attribute_type, attribute_value, source=None, displayed=None, params=None
    ):
        """Add a attribute to a Group/Indicator or Victim."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        if params is None:
            params = {}

        return self.tc_requests.add_attribute(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_type,
            attribute_value,
            source=source,
            displayed=displayed,
            owner=self.owner,
            params=params,
        )

    def update_attribute(
        self, attribute_value, attribute_id, source=None, displayed=None, params=None
    ):
        """Add a attribute to a Group/Indicator or Victim."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        if params is None:
            params = {}

        return self.tc_requests.update_attribute(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_value,
            attribute_id,
            source=source,
            displayed=displayed,
            owner=self.owner,
            params=params,
        )

    def delete_attribute(self, attribute_id):
        """Delete a attribute from a Group/Indicator or Victim."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.delete_attribute(
            self.api_type, self.api_branch, self.unique_id, attribute_id, owner=self.owner
        )

    def attribute_labels(self, attribute_id, params=None):
        """Get the security labels from a attribute."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        yield from self.tc_requests.attribute_labels(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_id,
            owner=self.owner,
            params=params,
        )

    def attribute_label(self, attribute_id, label, action='GET', params=None):
        """Get a security labels from a attribute."""
        if params is None:
            params = {}
        if not self.can_update():
            self._handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_attribute_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                label,
                owner=self.owner,
                params=params,
            )
        if action == 'DELETE':
            return self.tc_requests.delete_attribute_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                label,
                owner=self.owner,
            )

        self._handle_error(925, ['action', 'attribute_label', 'action', 'action', action])
        return None

    def add_attribute_label(self, attribute_id, label):
        """Add a security labels to a attribute."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.add_attribute_label(
            self.api_type, self.api_branch, self.unique_id, attribute_id, label, owner=self.owner
        )

    def can_create(self):
        """Determine if the object can be created."""
        return True

    def can_delete(self):
        """Determine if the object can be deleted."""
        return self.unique_id is not None

    def can_update(self):
        """Determine if the object can be updated."""
        return self.unique_id is not None

    @staticmethod
    def is_indicator():
        """Determine if the object is a Indicator."""
        return False

    @staticmethod
    def is_group():
        """Determine if the object is a Group."""
        return False

    @staticmethod
    def is_victim():
        """Determine if the object is a Victim."""
        return False

    @staticmethod
    def is_tag():
        """Determine if the object is a Tag."""
        return False

    @staticmethod
    def is_security_label():
        """Determine if the object is a Security Label."""
        return False

    @staticmethod
    def is_task():
        """Determine if the object is a Task."""
        return False

    @staticmethod
    def is_encoded(uri):
        """Determine if a uri is currently encoded"""
        uri = uri or ''

        return uri != unquote(uri)

    def fully_decode_uri(self, uri):
        """Decode a url till it is no longer encoded."""
        safety_valve = 0
        safety_valve_max_length = 10

        while self.is_encoded(uri):
            uri = unquote(uri)
            safety_valve += 1
            if safety_valve > safety_valve_max_length:
                break

        return uri

    def _set_unique_id(self, json_response):
        """Set the Unique Id given a json"""
        self.unique_id = json_response.get('id')

    def __str__(self):
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)



"""TcEx Framework Module"""

# standard library
import json
import logging
from collections import OrderedDict
from functools import cached_property
from pathlib import Path

from .install_json import InstallJson
from .model.tcex_json_model import TcexJsonModel
from .tcex_json_update import TcexJsonUpdate

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class TcexJson:
    """Config object for tcex.json configuration file"""

    def __init__(
        self,
        filename: str | None = None,
        path: Path | str | None = None,
        logger: logging.Logger | None = None,
    ):
        """Initialize instance properties."""
        filename = filename or 'tcex.json'
        path = Path(path or Path.cwd())
        self.log = logger or _logger

        # properties
        self.fqfn = path / filename
        self.ij = InstallJson(logger=self.log)

    @cached_property
    def contents(self) -> dict:
        """Return tcex.json file contents."""
        _contents = {}

        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    _contents = json.load(fh, object_pairs_hook=OrderedDict)
            except OSError:  # pragma: no cover
                self.log.exception(
                    f'feature=tcex-json, exception=failed-reading-file, filename={self.fqfn}'
                )
        else:  # pragma: no cover
            self.log.error(f'feature=tcex-json, exception=file-not-found, filename={self.fqfn}')

        return _contents

    @cached_property
    def model(self) -> TcexJsonModel:
        """Return the Install JSON model."""
        return TcexJsonModel(**self.contents)

    @property
    def update(self) -> TcexJsonUpdate:
        """Return InstallJsonUpdate instance."""
        return TcexJsonUpdate(tj=self)

    def write(self):
        """Write current data file."""
        data = self.model.json(exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True)
        with self.fqfn.open(mode='w') as fh:
            fh.write(f'{data}\n')



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class CaseAttributeFilter(FilterABC):
    """Filter Object for CaseAttributes"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.CASE_ATTRIBUTES.value

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Workflow ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case the workflow attribute is applied to.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the attribute was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def date_val(self, operator: Enum, date_val: Arrow | datetime | int | str):
        """Filter Date based on **dateVal** keyword.

        Args:
            operator: The operator enum for the filter.
            date_val: The date value of the attribute (only applies to certain types).
        """
        date_val = self.util.any_to_datetime(date_val).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateVal', operator, date_val, TqlType.STRING)

    def default(self, operator: Enum, default: bool):
        """Filter Default based on **default** keyword.

        Args:
            operator: The operator enum for the filter.
            default: A flag that is set by an attribute type configuration.
        """
        self._tql.add_filter('default', operator, default, TqlType.BOOLEAN)

    def displayed(self, operator: Enum, displayed: bool):
        """Filter Displayed based on **displayed** keyword.

        Args:
            operator: The operator enum for the filter.
            displayed: Whether or not the attribute is displayed on the item.
        """
        self._tql.add_filter('displayed', operator, displayed, TqlType.BOOLEAN)

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the attribute.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def int_val(self, operator: Enum, int_val: int | list):
        """Filter Integer Value based on **intVal** keyword.

        Args:
            operator: The operator enum for the filter.
            int_val: The integer value of the attribute (only applies to certain types).
        """
        if isinstance(int_val, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('intVal', operator, int_val, TqlType.INTEGER)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date the attribute was last modified in the system.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def max_size(self, operator: Enum, max_size: int | list):
        """Filter Max Size based on **maxSize** keyword.

        Args:
            operator: The operator enum for the filter.
            max_size: The max length of the attribute text.
        """
        if isinstance(max_size, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('maxSize', operator, max_size, TqlType.INTEGER)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the attribute.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the attribute.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def pinned(self, operator: Enum, pinned: bool):
        """Filter Pinned based on **pinned** keyword.

        Args:
            operator: The operator enum for the filter.
            pinned: Whether or not the attribute is pinned with importance.
        """
        self._tql.add_filter('pinned', operator, pinned, TqlType.BOOLEAN)

    def short_text(self, operator: Enum, short_text: list | str):
        """Filter Text based on **shortText** keyword.

        Args:
            operator: The operator enum for the filter.
            short_text: The short text of the attribute (only applies to certain types).
        """
        if isinstance(short_text, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('shortText', operator, short_text, TqlType.STRING)

    def source(self, operator: Enum, source: list | str):
        """Filter Source based on **source** keyword.

        Args:
            operator: The operator enum for the filter.
            source: The source text of the attribute.
        """
        if isinstance(source, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('source', operator, source, TqlType.STRING)

    def text(self, operator: Enum, text: list | str):
        """Filter Text based on **text** keyword.

        Args:
            operator: The operator enum for the filter.
            text: The text of the attribute (only applies to certain types).
        """
        if isinstance(text, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('text', operator, text, TqlType.STRING)

    def type(self, operator: Enum, type: int | list):  # noqa: A002
        """Filter Type ID based on **type** keyword.

        Args:
            operator: The operator enum for the filter.
            type: The ID of the attribute type.
        """
        if isinstance(type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('type', operator, type, TqlType.INTEGER)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter Type Name based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The name of the attribute type.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)

    def user(self, operator: Enum, user: list | str):
        """Filter User based on **user** keyword.

        Args:
            operator: The operator enum for the filter.
            user: The user who created the attribute.
        """
        if isinstance(user, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('user', operator, user, TqlType.STRING)



"""TcEx Framework Module"""

# standard library
import re

from .model.playbook_variable_model import PlaybookVariableModel


class BinaryVariable(bytes):
    """Bytes object with internal variable type field. Used when reading playbook variables"""

    _variable_type = 'Binary'


class StringVariable(str):
    """String object with internal variable type field. Used when reading playbook variables"""

    _variable_type = 'String'


class Variable:
    """TcEx Utilities Variables Class"""

    def contains_playbook_variable(self, key: str) -> bool:
        """Return True if provided key contains a properly formatted playbook variable."""
        return isinstance(key, str) and bool(re.search(self.variable_playbook_pattern, key))

    def contains_tc_variable(self, key: str):
        """Return True if provided key contains a properly formatted TC variable."""
        return isinstance(key, str) and bool(re.search(self.variable_tc_pattern, key))

    def get_playbook_variable_model(self, variable: str | None) -> PlaybookVariableModel | None:
        """Return data model of playbook variable (e.g., #App:1234:output!String)."""
        if variable is None:
            return None

        data = None
        variable = variable.strip()
        if re.match(self.variable_playbook_match, variable):
            var = re.search(self.variable_playbook_parse, variable)
            if var is not None:
                data = PlaybookVariableModel(**var.groupdict())
        return data

    def get_playbook_variable_type(self, variable: str) -> str:
        """Get variable type"""
        model = self.get_playbook_variable_model(variable)
        return 'String' if model is None else model.type

    def is_playbook_variable(self, key: str) -> bool:
        """Return True if provided key is a properly formatted playbook variable."""
        return isinstance(key, str) and bool(re.match(self.variable_playbook_match, key))

    def is_tc_variable(self, key: str) -> bool:
        """Return True if provided key is a properly formatted tc variable."""
        return isinstance(key, str) and bool(re.match(self.variable_tc_match, key))

    @property
    def variable_expansion_pattern(self) -> re.Pattern:
        """Regex pattern to match and parse a playbook or ThreatConnect variable.

        Playbook Variable : #App:334:example.service_input!String
        TC Variable       : &{TC:TEXT:4dc9202e-6945-4364-aa40-4b47655046d2}
        """
        return re.compile(
            # Origin:
            # PB-Variable: "#"
            # TC-Variable: "&"
            r'(?P<origin>#|&)'
            r'(?:\{)?'  # drop "{"
            # Provider:
            # PB-Variable: literal "App"
            # TC-Variable: provider (e.g. TC|Vault)
            r'(?P<provider>[A-Za-z]+):'
            # ID:
            # PB-Variable: Job ID (e.g., 334)
            # TC-Variable: One of (FILE|KEYCHAIN|TEXT)
            r'(?P<id>[\w]+):'
            # Lookup:
            # PB-Variable: variable id (e.g., API Token)
            # TC-Variable: variable id (e.g., 4dc9202e-6945-4364-aa40-4b47655046d2)
            r'(?P<lookup>[A-Za-z0-9_\.\-\[\]]+)'
            r'(?:\})?'  # drop "}"
            # Type
            # PB-Variable: variable type (e.g., String|StringArray)
            # TC-Variable: N/A
            # r'(?:!(?P<type>[A-Za-z0-9_-]+))?'
            r'(?:!(?P<type>StringArray|BinaryArray|KeyValueArray'
            r'|TCEntityArray|TCEnhancedEntityArray'
            r'|String|Binary|KeyValue|TCEntity|TCEnhancedEntity'
            r'|(?:(?!String)(?!Binary)(?!KeyValue)'
            r'(?!TCEntity)(?!TCEnhancedEntity)'
            r'[A-Za-z0-9_-]+)))?'  # variable type (custom)
        )

    @property
    def variable_playbook_array_types(self) -> list[str]:
        """Return list of standard playbook array variable types."""
        return [
            'BinaryArray',
            'KeyValueArray',
            'StringArray',
            'TCEntityArray',
            'TCEnhancedEntityArray',
        ]

    @property
    def variable_playbook_match(self) -> re.Pattern:
        """Return compiled re pattern for exact match of variable."""
        return re.compile(rf'^{self.variable_playbook_pattern}$')

    def variable_playbook_method_name(self, variable: str) -> str | None:
        """Convert variable name to a valid method name.

        #App:9876:string.operation!String -> string_operation_string
        """
        method_name = None
        if variable is not None:
            variable = variable.strip()
            if re.match(self.variable_playbook_match, variable):
                var = re.search(self.variable_playbook_parse, variable)
                if var is not None:
                    variable_name = var.group(3).replace('.', '_').lower()
                    variable_type = var.group(4).lower()
                    method_name = f'{variable_name}_{variable_type}'
        return method_name

    @property
    def variable_playbook_parse(self) -> re.Pattern:
        """Return compiled re pattern."""
        return re.compile(self.variable_playbook_pattern)

    @property
    def variable_playbook_pattern(self) -> str:
        """Regex pattern to match and parse a playbook variable.

        Parse this string -> #App:334:example.service_input!String
        """
        return (
            # App Type: one of the following types (APP|Trigger)
            r'#(?P<app_type>[A-Za-z]+)'
            # Job ID: the Id of the running job (e.g, 7979).
            r':(?P<job_id>[\d]+)'
            # Key: the variable key (e.g., api_token)
            r':(?P<key>[A-Za-z0-9_\.\-\[\]]+)'
            # Type: one of the following types
            #     (Binary|BinaryArray|KeyValue|KeyValueArray|
            #      String|StringArray|TCEntity|TCEntityArray|
            #      or <custom>)
            r'!(?P<type>StringArray|BinaryArray|KeyValueArray'
            r'|TCEntityArray|TCEnhancedEntityArray'
            r'|String|Binary|KeyValue|TCEntity|TCEnhancedEntity'
            r'|(?:(?!String)(?!Binary)(?!KeyValue)'
            r'(?!TCEntity)(?!TCEnhancedEntity)'
            r'[A-Za-z0-9_-]+))'  # variable type (custom)
        )

    @property
    def variable_playbook_single_types(self) -> list[str]:
        """Return list of standard playbook single variable types."""
        return [
            'Binary',
            'KeyValue',
            'String',
            'TCEntity',
            'TCEnhancedEntity',
        ]

    @property
    def variable_playbook_types(self) -> list[str]:
        """Return list of standard playbook variable types."""
        return self.variable_playbook_single_types + self.variable_playbook_array_types

    @property
    def variable_tc_match(self) -> re.Pattern:
        """Return regex pattern for tc variable EXACT match."""
        return re.compile(rf'^{self.variable_tc_pattern}$')

    @property
    def variable_tc_parse(self) -> re.Pattern:
        """Return regex pattern for tc variable search."""
        return re.compile(self.variable_tc_pattern)

    @property
    def variable_tc_pattern(self) -> str:
        """Return regex pattern for tc variable.

        Parse this string -> &{TC:TEXT:4dc9202e-6945-4364-aa40-4b47655046d2}
        """
        return (
            # Origin "&"
            r'(?:&)'
            r'(?:\{)'
            # Provider: who provides the variable (e.g. TC|Vault)
            r'(?P<provider>[A-Za-z]+):'
            # Type: one of (FILE|KEYCHAIN|TEXT)
            r'(?P<type>FILE|KEYCHAIN|TEXT):'
            # Key: variable id (e.g., 4dc9202e-6945-4364-aa40-4b47655046d2)
            r'(?P<key>[A-Za-z0-9_\.\-\[\]]+)'
            r'(?:\})'
        )



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel, ArtifactsModel
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.indicators.indicator import Indicator  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.notes.note import Note  # CIRCULAR-IMPORT


class Artifact(ObjectABC):
    """Artifacts Object.

    Args:
        associated_groups (Groups, kwargs): A list of Groups associated with this Artifact.
        associated_indicators (Indicators, kwargs): A list of Indicators associated with this
            Artifact.
        case_id (int, kwargs): The **case id** for the Artifact.
        case_xid (str, kwargs): The **case xid** for the Artifact.
        derived_link (bool, kwargs): Flag to specify if this artifact should be used for potentially
            associated cases or not.
        field_name (str, kwargs): The field name for the artifact.
        file_data (str, kwargs): Base64 encoded file attachment required only for certain artifact
            types.
        hash_code (str, kwargs): Hashcode of Artifact of type File.
        notes (Notes, kwargs): A list of Notes corresponding to the Artifact.
        source (str, kwargs): The **source** for the Artifact.
        summary (str, kwargs): The **summary** for the Artifact.
        task_id (int, kwargs): The ID of the task which the Artifact references.
        task_xid (str, kwargs): The XID of the task which the Artifact references.
        type (str, kwargs): The **type** for the Artifact.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: ArtifactModel = ArtifactModel(**kwargs)
        self._nested_field_name = 'artifacts'
        self._nested_filter = 'has_artifact'
        self.type_ = 'Artifact'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.ARTIFACTS.value

    @property
    def model(self) -> ArtifactModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | ArtifactModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.summary}

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    @property
    def associated_indicators(self) -> Generator['Indicator', None, None]:
        """Yield Indicator from Indicators."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator import Indicators

        yield from self._iterate_over_sublist(Indicators)  # type: ignore

    @property
    def notes(self) -> Generator['Note', None, None]:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            ex_msg = 'Invalid type passed in to stage_associated_group'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_groups.data.append(data)  # type: ignore

    def stage_associated_indicator(self, data: dict | ObjectABC | IndicatorModel):
        """Stage indicator on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = IndicatorModel(**data)

        if not isinstance(data, IndicatorModel):
            ex_msg = 'Invalid type passed in to stage_associated_indicator'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_indicators.data.append(data)  # type: ignore

    def stage_note(self, data: dict | ObjectABC | NoteModel):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            ex_msg = 'Invalid type passed in to stage_note'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.notes.data.append(data)  # type: ignore


class Artifacts(ObjectCollectionABC):
    """Artifacts Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = ArtifactsModel(**kwargs)
        self.type_ = 'artifacts'

    def __iter__(self) -> Iterator[Artifact]:
        """Return CM objects."""
        return self.iterate(base_class=Artifact)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.ARTIFACTS.value

    @property
    def filter(self) -> ArtifactFilter:
        """Return the type specific filter object."""
        return ArtifactFilter(self.tql)



"""TcEx Framework Module"""

# standard library
import time
from collections.abc import Callable

# third-party
from requests import PreparedRequest, auth

from ...input.field_type.sensitive import Sensitive


class TokenAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, tc_token: Callable[..., Sensitive | str] | Sensitive | str):
        """Initialize the Class properties."""
        auth.AuthBase.__init__(self)
        self.tc_token = tc_token

    def _token_header(self):
        """Return HMAC Authorization header value."""
        _token = self.tc_token
        if callable(self.tc_token):
            # Callable - A callable method is provided that will return the token as a plain
            #     string. The callable will have to handle token renewal.
            _token = self.tc_token()

        if isinstance(_token, Sensitive):
            # Sensitive - A sensitive string type was passed. Likely no support for renewal.
            _token = _token.value

        # Return formatted token
        return f'TC-Token {_token}'

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the authorization headers to the request."""
        timestamp = int(time.time())

        # Add required headers to auth.
        r.headers['Authorization'] = self._token_header()
        r.headers['Timestamp'] = str(timestamp)
        return r



"""TcEx Framework Module"""

# first-party
from tcex.input.model.common_model import CommonModel
from tcex.input.model.organization_model import OrganizationModel


class AppOrganizationModel(CommonModel, OrganizationModel):
    """Model Definition"""



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class ArtifactModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Artifact Model',
    validate_assignment=True,
):
    """Artifact Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=True)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    analytics_priority: str | None = Field(
        None,
        allow_mutation=False,
        description='The **analytics priority** for the Artifact.',
        read_only=True,
        title='analyticsPriority',
    )
    analytics_priority_level: int | None = Field(
        None,
        allow_mutation=False,
        description='The **analytics priority level** for the Artifact.',
        read_only=True,
        title='analyticsPriorityLevel',
    )
    analytics_score: int | None = Field(
        None,
        allow_mutation=False,
        description='The **analytics score** for the Artifact.',
        read_only=True,
        title='analyticsScore',
    )
    analytics_status: str | None = Field(
        None,
        allow_mutation=False,
        description='The **analytics status** for the Artifact.',
        read_only=True,
        title='analyticsStatus',
    )
    analytics_type: str | None = Field(
        None,
        allow_mutation=False,
        description='The **analytics type** for the Artifact.',
        read_only=True,
        title='analyticsType',
    )
    artifact_type: 'ArtifactTypeModel' = Field(
        None,
        allow_mutation=False,
        description='The **artifact type** for the Artifact.',
        read_only=True,
        title='artifactType',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of Groups associated with this Artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: 'IndicatorsModel' = Field(
        None,
        description='A list of Indicators associated with this Artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    case_id: int | None = Field(
        None,
        description='The **case id** for the Artifact.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseXid',
        title='caseId',
    )
    case_xid: str | None = Field(
        None,
        description='The **case xid** for the Artifact.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseId',
        title='caseXid',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The **date added** for the Artifact.',
        read_only=True,
        title='dateAdded',
    )
    derived_link: bool = Field(
        None,
        description=(
            'Flag to specify if this artifact should be used for potentially associated cases or '
            'not.'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='derivedLink',
    )
    field_name: str | None = Field(
        None,
        description='The field name for the artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fieldName',
    )
    file_data: str | None = Field(
        None,
        description='Base64 encoded file attachment required only for certain artifact types.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileData',
    )
    hash_code: str | None = Field(
        None,
        description='Hashcode of Artifact of type File.',
        methods=['POST'],
        read_only=False,
        title='hashCode',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    intel_type: str | None = Field(
        None,
        allow_mutation=False,
        description='The **intel type** for the Artifact.',
        read_only=True,
        title='intelType',
    )
    links: dict | None = Field(
        None,
        allow_mutation=False,
        description='The **links** for the Artifact.',
        read_only=True,
        title='links',
    )
    notes: 'NotesModel' = Field(
        None,
        description='A list of Notes corresponding to the Artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='notes',
    )
    parent_case: 'CaseModel' = Field(
        None,
        allow_mutation=False,
        description='The **parent case** for the Artifact.',
        read_only=True,
        title='parentCase',
    )
    source: str | None = Field(
        None,
        description='The **source** for the Artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='source',
    )
    summary: str | None = Field(
        None,
        description='The **summary** for the Artifact.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='summary',
    )
    task: 'TaskModel' = Field(
        None,
        allow_mutation=False,
        description='The **task** for the Artifact.',
        read_only=True,
        title='task',
    )
    task_id: int | None = Field(
        None,
        description='The ID of the task which the Artifact references.',
        methods=['POST'],
        read_only=False,
        title='taskId',
    )
    task_xid: str | None = Field(
        None,
        description='The XID of the task which the Artifact references.',
        methods=['POST'],
        read_only=False,
        title='taskXid',
    )
    type: str | None = Field(
        None,
        description='The **type** for the Artifact.',
        methods=['POST'],
        read_only=False,
        title='type',
    )

    @validator('artifact_type', always=True, pre=True)
    @classmethod
    def _validate_artifact_type(cls, v):
        if not v:
            return ArtifactTypeModel()  # type: ignore
        return v

    @validator('parent_case', always=True, pre=True)
    @classmethod
    def _validate_case(cls, v):
        if not v:
            return CaseModel()  # type: ignore
        return v

    @validator('associated_groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('associated_indicators', always=True, pre=True)
    @classmethod
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('notes', always=True, pre=True)
    @classmethod
    def _validate_notes(cls, v):
        if not v:
            return NotesModel()  # type: ignore
        return v

    @validator('task', always=True, pre=True)
    @classmethod
    def _validate_task(cls, v):
        if not v:
            return TaskModel()  # type: ignore
        return v


class ArtifactDataModel(
    BaseModel,
    title='Artifact Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Artifacts Data Model"""

    data: list[ArtifactModel] | None = Field(
        [],
        description='The data for the Artifacts.',
        methods=['POST', 'PUT'],
        title='data',
    )


class ArtifactsModel(
    BaseModel,
    title='Artifacts Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Artifacts Model"""

    _mode_support = PrivateAttr(default=True)

    data: list[ArtifactModel] | None = Field(
        [],
        description='The data for the Artifacts.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifact_types.artifact_type_model import ArtifactTypeModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.notes.note_model import NotesModel
from tcex.api.tc.v3.tasks.task_model import TaskModel

# add forward references
ArtifactDataModel.update_forward_refs()
ArtifactModel.update_forward_refs()
ArtifactsModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.api.tc.v2.batch.batch import Batch
from tcex.api.tc.v2.batch.batch_submit import BatchSubmit
from tcex.api.tc.v2.batch.batch_writer import BatchWriter
from tcex.api.tc.v2.datastore.cache import Cache
from tcex.api.tc.v2.datastore.datastore import DataStore
from tcex.api.tc.v2.metric.metric import Metric
from tcex.api.tc.v2.notification.notification import Notification
from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence
from tcex.input.input import Input  # TYPE-CHECKING


class V2:
    """API -> TC -> V2

    Args:
        inputs: An instance of the Input class.
        session_tc: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, inputs: Input, session_tc: Session):
        """Initialize instance properties."""
        self.inputs = inputs
        self.session_tc = session_tc

    def batch(
        self,
        owner: str,
        action: str = 'Create',
        attribute_write_type: str = 'Replace',
        halt_on_error: bool = False,
        playbook_triggers_enabled: bool = False,
        tag_write_type: str = 'Replace',
        security_label_write_type: str = 'Replace',
    ) -> Batch:
        """Return instance of Batch

        Args:
            owner: The ThreatConnect owner for Batch action.
            action: Action for the batch job ['Create', 'Delete'].
            attribute_write_type: Write type for TI attributes ['Append', 'Replace'].
            halt_on_error: If True any batch error will halt the batch job.
            playbook_triggers_enabled: Deprecated input, will not be used.
            security_label_write_type: Write type for labels ['Append', 'Replace'].
            tag_write_type: Write type for tags ['Append', 'Replace'].
        """
        return Batch(
            self.inputs,
            self.session_tc,
            owner,
            action,
            attribute_write_type,
            halt_on_error,
            playbook_triggers_enabled,
            tag_write_type,
            security_label_write_type,
        )

    def batch_submit(
        self,
        owner: str,
        action: str = 'Create',
        attribute_write_type: str = 'Replace',
        halt_on_error: bool = False,
        playbook_triggers_enabled: bool = False,
        tag_write_type: str = 'Replace',
        security_label_write_type: str = 'Replace',
    ) -> BatchSubmit:
        """Return instance of Batch

        Args:
            owner: The ThreatConnect owner for Batch action.
            action: Action for the batch job ['Create', 'Delete'].
            attribute_write_type: Write type for TI attributes ['Append', 'Replace'].
            halt_on_error: If True any batch error will halt the batch job.
            playbook_triggers_enabled: Deprecated input, will not be used.
            security_label_write_type: Write type for labels ['Append', 'Replace'].
            tag_write_type: Write type for tags ['Append', 'Replace'].
        """
        return BatchSubmit(
            self.inputs,
            self.session_tc,
            owner,
            action,
            attribute_write_type,
            halt_on_error,
            playbook_triggers_enabled,
            tag_write_type,
            security_label_write_type,
        )

    def batch_writer(self, output_dir: str, **kwargs) -> BatchWriter:
        """Return instance of Batch

        Args:
            output_dir: Deprecated input, will not be used.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            output_extension (str): Append this extension to output files.
            write_callback (Callable): A callback method to call when a batch json file is
                written. The callback will be passed the fully qualified name of the written file.
            write_callback_kwargs (dict): Additional values to send to callback method.
        """
        return BatchWriter(self.inputs, self.session_tc, output_dir, **kwargs)

    def cache(
        self,
        domain: str,
        data_type: str,
        ttl_seconds: int | None = None,
        mapping: dict | None = None,
    ) -> Cache:
        """Get instance of the Cache module.

        Args:
            domain: The domain can be either "system", "organization", or "local". When using
                "organization" the data store can be accessed by any Application in the entire org,
                while "local" access is restricted to the App writing the data. The "system" option
                should not be used in almost all cases.
            data_type: The data type descriptor (e.g., tc:whois:cache).
            ttl_seconds: The number of seconds the cache is valid.
            mapping: Advanced - The datastore mapping if required.
        """
        return Cache(self.session_tc, domain, data_type, ttl_seconds, mapping)

    def datastore(self, domain: str, data_type: str, mapping: dict | None = None) -> DataStore:
        """Return Datastore Module.

        Args:
            domain: The domain can be either "system", "organization", or "local". When using
                "organization" the data store can be accessed by any Application in the entire org,
                while "local" access is restricted to the App writing the data. The "system" option
                should not be used in almost all cases.
            data_type: The data type descriptor (e.g., tc:whois:cache).
            mapping: ElasticSearch mapping data.
        """
        return DataStore(
            session_tc=self.session_tc, domain=domain, data_type=data_type, mapping=mapping
        )

    def metric(
        self,
        name: str,
        description: str,
        data_type: str,
        interval: str,
        keyed: bool = False,
    ) -> Metric:
        """Get instance of the Metrics module.

        Args:
            name: The name for the metric.
            description: The description of the metric.
            data_type: The type of metric: Sum, Count, Min, Max, First, Last, and Average.
            interval: The metric interval: Hourly, Daily, Weekly, Monthly, and Yearly.
            keyed: Indicates whether the data will have a keyed value.
        """
        return Metric(
            session_tc=self.session_tc,
            name=name,
            description=description,
            data_type=data_type,
            interval=interval,
            keyed=keyed,
        )

    @property
    def notification(self) -> Notification:
        """Get instance of the Notification module."""
        return Notification(self.session_tc)

    @property
    def ti(self) -> ThreatIntelligence:
        """Get instance of the Notification module."""
        return ThreatIntelligence(self.session_tc)



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class CaseAttributeModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='CaseAttribute Model',
    validate_assignment=True,
):
    """Case_Attribute Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    case_id: int | None = Field(
        None,
        description='Case associated with attribute.',
        methods=['POST'],
        read_only=False,
        title='caseId',
    )
    created_by: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The **created by** for the Case_Attribute.',
        read_only=True,
        title='createdBy',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    default: bool = Field(
        None,
        description=(
            'A flag indicating that this is the default attribute of its type within the object. '
            'Only applies to certain attribute and data types.'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='default',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    last_modified: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Attribute was last modified.',
        read_only=True,
        title='lastModified',
    )
    pinned: bool = Field(
        None,
        description='A flag indicating that the attribute has been noted for importance.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='pinned',
    )
    security_labels: 'SecurityLabelsModel' = Field(
        None,
        description=(
            'A list of Security Labels corresponding to the Intel item (NOTE: Setting this '
            'parameter will replace any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityLabels',
    )
    source: str | None = Field(
        None,
        description='The attribute source.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='source',
    )
    type: str | None = Field(
        None,
        description='The attribute type.',
        methods=['POST'],
        read_only=False,
        title='type',
    )
    value: str | None = Field(
        None,
        description='The attribute value.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='value',
    )

    @validator('security_labels', always=True, pre=True)
    @classmethod
    def _validate_security_labels(cls, v):
        if not v:
            return SecurityLabelsModel()  # type: ignore
        return v

    @validator('created_by', always=True, pre=True)
    @classmethod
    def _validate_user(cls, v):
        if not v:
            return UserModel()  # type: ignore
        return v


class CaseAttributeDataModel(
    BaseModel,
    title='CaseAttribute Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Case_Attributes Data Model"""

    data: list[CaseAttributeModel] | None = Field(
        [],
        description='The data for the CaseAttributes.',
        methods=['POST', 'PUT'],
        title='data',
    )


class CaseAttributesModel(
    BaseModel,
    title='CaseAttributes Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Case_Attributes Model"""

    _mode_support = PrivateAttr(default=True)

    data: list[CaseAttributeModel] | None = Field(
        [],
        description='The data for the CaseAttributes.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelsModel

# add forward references
CaseAttributeDataModel.update_forward_refs()
CaseAttributeModel.update_forward_refs()
CaseAttributesModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class OwnerFilter(FilterABC):
    """Filter Object for Owners"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.OWNERS.value

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the Community Membership.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def owner_id(self, operator: Enum, owner_id: int | list):
        """Filter Owner ID based on **ownerId** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_id: The ID of the Owner.
        """
        if isinstance(owner_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerId', operator, owner_id, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The name of the Owner.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def perm_apps(self, operator: Enum, perm_apps: list | str):
        """Filter Apps Permission based on **permApps** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_apps: The User's Apps permission in the Owner.
        """
        if isinstance(perm_apps, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permApps', operator, perm_apps, TqlType.STRING)

    def perm_artifact(self, operator: Enum, perm_artifact: list | str):
        """Filter Artifact Permission based on **permArtifact** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_artifact: The User's Artifact permission in the Owner.
        """
        if isinstance(perm_artifact, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permArtifact', operator, perm_artifact, TqlType.STRING)

    def perm_attribute(self, operator: Enum, perm_attribute: list | str):
        """Filter Attribute Permission based on **permAttribute** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_attribute: The User's Attribute permission in the Owner.
        """
        if isinstance(perm_attribute, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permAttribute', operator, perm_attribute, TqlType.STRING)

    def perm_attribute_type(self, operator: Enum, perm_attribute_type: list | str):
        """Filter AttributeType Permission based on **permAttributeType** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_attribute_type: The User's AttributeType permission in the Owner.
        """
        if isinstance(perm_attribute_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permAttributeType', operator, perm_attribute_type, TqlType.STRING)

    def perm_case_tag(self, operator: Enum, perm_case_tag: list | str):
        """Filter CaseTag Permission based on **permCaseTag** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_case_tag: The User's CaseTag permission in the Owner.
        """
        if isinstance(perm_case_tag, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permCaseTag', operator, perm_case_tag, TqlType.STRING)

    def perm_comment(self, operator: Enum, perm_comment: list | str):
        """Filter Comment Permission based on **permComment** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_comment: The User's Comment permission in the Owner.
        """
        if isinstance(perm_comment, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permComment', operator, perm_comment, TqlType.STRING)

    def perm_copy_data(self, operator: Enum, perm_copy_data: list | str):
        """Filter CopyData Permission based on **permCopyData** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_copy_data: The User's CopyData permission in the Owner.
        """
        if isinstance(perm_copy_data, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permCopyData', operator, perm_copy_data, TqlType.STRING)

    def perm_group(self, operator: Enum, perm_group: list | str):
        """Filter Group Permission based on **permGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_group: The User's Group permission in the Owner.
        """
        if isinstance(perm_group, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permGroup', operator, perm_group, TqlType.STRING)

    def perm_indicator(self, operator: Enum, perm_indicator: list | str):
        """Filter Indicator Permission based on **permIndicator** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_indicator: The User's Indicator permission in the Owner.
        """
        if isinstance(perm_indicator, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permIndicator', operator, perm_indicator, TqlType.STRING)

    def perm_invite(self, operator: Enum, perm_invite: list | str):
        """Filter Invite Permission based on **permInvite** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_invite: The User's Invite permission in the Owner.
        """
        if isinstance(perm_invite, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permInvite', operator, perm_invite, TqlType.STRING)

    def perm_members(self, operator: Enum, perm_members: list | str):
        """Filter Members Permission based on **permMembers** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_members: The User's Members permission in the Owner.
        """
        if isinstance(perm_members, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permMembers', operator, perm_members, TqlType.STRING)

    def perm_playbooks(self, operator: Enum, perm_playbooks: list | str):
        """Filter Playbooks Permission based on **permPlaybooks** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_playbooks: The User's Playbooks permission in the Owner.
        """
        if isinstance(perm_playbooks, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permPlaybooks', operator, perm_playbooks, TqlType.STRING)

    def perm_playbooks_execute(self, operator: Enum, perm_playbooks_execute: list | str):
        """Filter PlaybooksExecute Permission based on **permPlaybooksExecute** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_playbooks_execute: The User's PlaybooksExecute permission in the Owner.
        """
        if isinstance(perm_playbooks_execute, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter(
            'permPlaybooksExecute', operator, perm_playbooks_execute, TqlType.STRING
        )

    def perm_post(self, operator: Enum, perm_post: list | str):
        """Filter Post Permission based on **permPost** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_post: The User's Post permission in the Owner.
        """
        if isinstance(perm_post, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permPost', operator, perm_post, TqlType.STRING)

    def perm_publish(self, operator: Enum, perm_publish: list | str):
        """Filter Publish Permission based on **permPublish** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_publish: The User's Publish permission in the Owner.
        """
        if isinstance(perm_publish, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permPublish', operator, perm_publish, TqlType.STRING)

    def perm_security_label(self, operator: Enum, perm_security_label: list | str):
        """Filter SecurityLabel Permission based on **permSecurityLabel** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_security_label: The User's SecurityLabel permission in the Owner.
        """
        if isinstance(perm_security_label, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permSecurityLabel', operator, perm_security_label, TqlType.STRING)

    def perm_settings(self, operator: Enum, perm_settings: list | str):
        """Filter Settings Permission based on **permSettings** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_settings: The User's Settings permission in the Owner.
        """
        if isinstance(perm_settings, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permSettings', operator, perm_settings, TqlType.STRING)

    def perm_tag(self, operator: Enum, perm_tag: list | str):
        """Filter Tag Permission based on **permTag** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_tag: The User's Tag permission in the Owner.
        """
        if isinstance(perm_tag, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permTag', operator, perm_tag, TqlType.STRING)

    def perm_task(self, operator: Enum, perm_task: list | str):
        """Filter Task Permission based on **permTask** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_task: The User's Task permission in the Owner.
        """
        if isinstance(perm_task, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permTask', operator, perm_task, TqlType.STRING)

    def perm_timeline(self, operator: Enum, perm_timeline: list | str):
        """Filter Timeline Permission based on **permTimeline** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_timeline: The User's Timeline permission in the Owner.
        """
        if isinstance(perm_timeline, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permTimeline', operator, perm_timeline, TqlType.STRING)

    def perm_track(self, operator: Enum, perm_track: list | str):
        """Filter Track Permission based on **permTrack** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_track: The User's Track permission in the Owner.
        """
        if isinstance(perm_track, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permTrack', operator, perm_track, TqlType.STRING)

    def perm_users(self, operator: Enum, perm_users: list | str):
        """Filter Users Permission based on **permUsers** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_users: The User's Users permission in the Owner.
        """
        if isinstance(perm_users, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permUsers', operator, perm_users, TqlType.STRING)

    def perm_victim(self, operator: Enum, perm_victim: list | str):
        """Filter Victim Permission based on **permVictim** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_victim: The User's Victim permission in the Owner.
        """
        if isinstance(perm_victim, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('permVictim', operator, perm_victim, TqlType.STRING)

    def perm_workflow_template(self, operator: Enum, perm_workflow_template: list | str):
        """Filter WorkflowTemplate Permission based on **permWorkflowTemplate** keyword.

        Args:
            operator: The operator enum for the filter.
            perm_workflow_template: The User's WorkflowTemplate permission in the Owner.
        """
        if isinstance(perm_workflow_template, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter(
            'permWorkflowTemplate', operator, perm_workflow_template, TqlType.STRING
        )

    def user_id(self, operator: Enum, user_id: int | list):
        """Filter User ID based on **userId** keyword.

        Args:
            operator: The operator enum for the filter.
            user_id: The ID of the user.
        """
        if isinstance(user_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('userId', operator, user_id, TqlType.INTEGER)
